//
//  Bridge.js
//  FolioReaderKit
//
//  Created by Heberti Almeida on 06/05/15.
//  Copyright (c) 2015 Folio Reader. All rights reserved.
//

var thisHighlight;
var audioMarkClass;
var wordsPerMinute = 180;

var times = function() {
    var timing = performance.timing;
    var loadTime = timing.loadEventEnd - timing.navigationStart;//过早获取时,loadEventEnd有时会是0
    if(loadTime <= 0) {
    // 未加载完，延迟200ms后继续times方法，直到成功
        setTimeout(function(){
            times();
        }, 200);
        return;
    }
    var readyStart = timing.fetchStart - timing.navigationStart;
    var redirectTime = timing.redirectEnd  - timing.redirectStart;
    var appcacheTime = timing.domainLookupStart  - timing.fetchStart;
    var unloadEventTime = timing.unloadEventEnd - timing.unloadEventStart;
    var lookupDomainTime = timing.domainLookupEnd - timing.domainLookupStart;
    var connectTime = timing.connectEnd - timing.connectStart;
    var requestTime = timing.responseEnd - timing.requestStart;
    var initDomTreeTime = timing.domInteractive - timing.responseEnd;
    var domReadyTime = timing.domComplete - timing.domInteractive; //过早获取时,domComplete有时会是0
    var loadEventTime = timing.loadEventEnd - timing.loadEventStart;

    // 为console.table方法准备对象，包含耗时的描述和消耗的时间
    var allTimes = [
        { "描述": "准备新页面时间耗时", "时间(ms)": readyStart },
        { "描述": "redirect 重定向耗时", "时间(ms)": redirectTime },
        { "描述": "Appcache 耗时", "时间(ms)": appcacheTime },
        { "描述": "unload 前文档耗时", "时间(ms)": unloadEventTime },
        { "描述": "DNS 查询耗时", "时间(ms)": lookupDomainTime },
        { "描述": "TCP连接耗时", "时间(ms)": connectTime },
        { "描述": "request请求耗时", "时间(ms)": requestTime },
        { "描述": "请求完毕至DOM加载", "时间(ms)": initDomTreeTime },
        { "描述": "解释dom树耗时", "时间(ms)": domReadyTime },
        { "描述": "load事件耗时", "时间(ms)": loadEventTime },
        { "描述": "从开始至load总耗时", "时间(ms)": loadTime }
    ];
    console.log(allTimes)
}

function getQueryString(name) {
          var reg = new RegExp('(^|&)' + name + '=([^&]*)(&|$)', 'i');
          var r = window.location.search.substr(1).match(reg);

          if (r != null) {
             return unescape(r[2]);
          }
          return null;
}


document.addEventListener("click", function(event) {
    //console.log(event)
    if(event && event.toElement) {
        if(event.toElement.hash){
            var hash = event.toElement.hash.replace('#','');
            scrollToSpan(true, hash)
        }
    }
});

function load(event){
   //var direction == "VERTICAL"
   //times()
   var direction = getQueryString('direction')
   var percent = parseFloat(getQueryString('percent') || '0')
   if(direction == 'HORIZONTAL'){
      initHorizontalDirection(percent)
   }else if(direction == 'VERTICAL') {
       var ratio = window.devicePixelRatio
       var scrollHeight = document.documentElement.scrollHeight
       var clientHeight = document.documentElement.clientHeight
       FolioPageFragment.loadingFinished(scrollHeight*ratio, clientHeight*ratio)
   }
   //times()
}

window.onload = load

function onchange(event){
  console.log('onchange')
  console.log(event)
}

window.onchange = onchange

// Generate a GUID
function guid() {
    function s4() {
        return Math.floor((1 + Math.random()) * 0x10000).toString(16).substring(1);
    }
    var guid = s4() + s4() + '-' + s4() + '-' + s4() + '-' + s4() + '-' + s4() + s4() + s4();
    return guid.toUpperCase();
}

// Get All HTML
function getHTML() {
    Highlight.getHtmlAndSaveHighlight(document.documentElement.outerHTML);
    //return document.documentElement.outerHTML;
}

// Class manipulation
function hasClass(ele,cls) {
  return !!ele.className.match(new RegExp('(\\s|^)'+cls+'(\\s|$)'));
}

function addClass(ele,cls) {
  if (!hasClass(ele,cls)) ele.className += " "+cls;
}

function removeClass(ele,cls) {
  if (hasClass(ele,cls)) {
    var reg = new RegExp('(\\s|^)'+cls+'(\\s|$)');
    ele.className=ele.className.replace(reg,' ');
  }
}

function removeClassArray(ele,array){
  for(var i = 0; i < array.length; ++i){
    removeClass(ele, array[i])
  }
}

// Toggle night mode
function nightMode(enable) {
    var elm = document.documentElement;
    if(enable) {
        addClass(elm, "nightMode");
    } else {
        removeClass(elm, "nightMode");
    }
}

// Font name class
function setFontName(cls,clsAry) {
    var elm = document.documentElement;

    if(cls === '') {
       removeClassArray(elm, clsAry)
       onFontLoadFinish()
       return;
    }
    var font = new FontFaceObserver(cls)
    font.load().then(function () {
      removeClassArray(elm, clsAry)
      addClass(elm, cls);
      onFontLoadFinish()
    }, function () {
       removeClassArray(elm, clsAry)
      console.log('Font is not available');
    });
}

function onFontLoadFinish(){
    var ratio = window.devicePixelRatio
    var scrollHeight = document.documentElement.scrollHeight
    var clientHeight = document.documentElement.clientHeight
    var scrollWidth = document.documentElement.scrollWidth
    var clientWidth = document.documentElement.clientWidth
    FolioPageFragment.onFontLoadFinish(scrollWidth*ratio, scrollHeight*ratio, clientWidth*ratio, clientHeight*ratio)
}

function initFontFamily(clsAry){
   for(var i = 0 ; i < clsAry.length; ++i) {
      var font = new FontFaceObserver(clsAry[i])
          font.load().then(function () {
          }, function () {
             removeClassArray(elm, clsAry)
            console.log('Font is not available');
          });
   }
}

// Set font size
function setFontSize(cls,clsAry) {
    var elm = document.documentElement;
    removeClassArray(elm, clsAry)
    addClass(elm, cls);
    //window.location.reload()
}


function setBackground(bgName,clsAry){
    var elm = document.documentElement;
    removeClass(elm, "nightMode");
    removeClassArray(elm, clsAry)
    addClass(elm, bgName);
}

// Menu colors
function setHighlightStyle(style) {
    Highlight.getUpdatedHighlightId(thisHighlight.id, style);
}

function removeThisHighlight() {
    return thisHighlight.id;
}

function removeHighlightById(elmId) {
    var elm = document.getElementById(elmId);
    elm.outerHTML = elm.innerHTML;
    return elm.id;
}

function getHighlightContent() {
    return thisHighlight.textContent
}

function getBodyText() {
    return document.body.innerText;
}

// Method that returns only selected text plain
var getSelectedText = function() {
    return window.getSelection().toString();
}

// Method that gets the Rect of current selected text
// and returns in a JSON format
var getRectForSelectedText = function(elm) {
    if (typeof elm === "undefined") elm = window.getSelection().getRangeAt(0);
    var rect = elm.getBoundingClientRect();
    return "{{" + rect.left + "," + rect.top + "}, {" + rect.width + "," + rect.height + "}}";
}

// Method that call that a hightlight was clicked
// with URL scheme and rect informations
var callHighlightURL = function(elm) {
    event.stopPropagation();
    var URLBase = "highlight://";
    var currentHighlightRect = getRectForSelectedText(elm);
    thisHighlight = elm;

    window.location = URLBase + encodeURIComponent(currentHighlightRect);
}

// Reading time
function getReadingTime() {
    var text = document.body.innerText;
    var totalWords = text.trim().split(/\s+/g).length;
    var wordsPerSecond = wordsPerMinute / 60; //define words per second based on words per minute
    var totalReadingTimeSeconds = totalWords / wordsPerSecond; //define total reading time in seconds
    var readingTimeMinutes = Math.round(totalReadingTimeSeconds / 60);

    return readingTimeMinutes;
}

/**
 Get Vertical or Horizontal paged #anchor offset
 */
var getAnchorOffset = function(target, horizontal) {
    var elem = document.getElementById(target);

    if (!elem) {
        elem = document.getElementsByName(target)[0];
    }

    if (horizontal) {
        return document.body.clientWidth * Math.floor(elem.offsetTop / window.innerHeight);
    }

    return elem.offsetTop;
}

function scrollAnchor(id) {
    window.location.hash = id;
}

function findElementWithID(node) {
    if( !node || node.tagName == "BODY")
        return null
    else if( node.id )
        return node
    else
        return findElementWithID(node)
}

function findElementWithIDInView() {

    if(audioMarkClass) {
        // attempt to find an existing "audio mark"
        var el = document.querySelector("."+audioMarkClass)

        // if that existing audio mark exists and is in view, use it
        if( el && el.offsetTop > document.body.scrollTop && el.offsetTop < (window.innerHeight + document.body.scrollTop))
            return el
    }

    // @NOTE: is `span` too limiting?
    var els = document.querySelectorAll("span[id]")

    for(indx in els) {
        var element = els[indx];

        // Horizontal scroll
        if (document.body.scrollTop == 0) {
            var elLeft = document.body.clientWidth * Math.floor(element.offsetTop / window.innerHeight);
            // document.body.scrollLeft = elLeft;

            if (elLeft == document.body.scrollLeft) {
                return element;
            }

        // Vertical
        } else if(element.offsetTop > document.body.scrollTop) {
            return element;
        }
    }

    return null
}


/**
 Play Audio - called by native UIMenuController when a user selects a bit of text and presses "play"
 */
function playAudio() {
    var sel = getSelection();
    var node = null;

    // user selected text? start playing from the selected node
    if (sel.toString() != "") {
        node = sel.anchorNode ? findElementWithID(sel.anchorNode.parentNode) : null;

    // find the first ID'd element that is within view (it will
    } else {
        node = findElementWithIDInView()
    }

    playAudioFragmentID(node ? node.id : null)
}


/**
 Play Audio Fragment ID - tells page controller to begin playing audio from the following ID
 */
function playAudioFragmentID(fragmentID) {
    var URLBase = "play-audio://";
    window.location = URLBase + (fragmentID?encodeURIComponent(fragmentID):"")
}

function bodyOrHtml() {
    if ('scrollingElement' in document) {
        return document.scrollingElement;
    }
    // Fallback for legacy browsers
    if (navigator.userAgent.indexOf('WebKit') != -1) {
        return document.body;
    }
    return document.documentElement;
}

function scrollToElement(element) {


    var scrollingElement = bodyOrHtml();

    if (FolioPageFragment.getDirection() == "VERTICAL") {

        var topDistraction = FolioPageFragment.getTopDistraction();
        var pageTop = scrollingElement.scrollTop + topDistraction;
        var pageBottom = scrollingElement.scrollTop + document.documentElement.clientHeight
                            - FolioPageFragment.getBottomDistraction();

        var elementTop = element.offsetTop - 20;
        elementTop = elementTop < 0 ? 0 : elementTop;
        var elementBottom = element.offsetTop + element.offsetHeight + 20;
        var needToScroll = (elementTop < pageTop || elementBottom > pageBottom);
        if (needToScroll) {
            var newScrollTop = elementTop - topDistraction;
            newScrollTop = newScrollTop < 0 ? 0 : newScrollTop;
            //console.log("-> Scrolled to = " + newScrollTop);
            scrollingElement.scrollTop = newScrollTop;
        }

    } else if (FolioPageFragment.getDirection() == "HORIZONTAL") {
        var clientWidth = document.documentElement.clientWidth;
        var pageIndex = Math.floor(element.offsetLeft / clientWidth);
        var newScrollLeft = clientWidth * pageIndex;
        //console.log("-> newScrollLeft = " + newScrollLeft);
        scrollingElement.scrollLeft = newScrollLeft;
        WebViewPager.setCurrentPage(pageIndex);
    }

    return element;
}

/**
 Remove All Classes - removes the given class from all elements in the DOM
 */
function removeAllClasses(className) {
    var els = document.body.getElementsByClassName(className)
    if( els.length > 0 )
    for( i = 0; i <= els.length; i++) {
        els[i].classList.remove(className);
    }
}

//For testing purpose only
function sleep(seconds)
{
  var e = new Date().getTime() + (seconds * 1000);
  while (new Date().getTime() <= e) {}
}

var searchResults = [];
var lastSearchQuery = null;
var testCounter = 0;
var searchResultsInvisible = true;

// Testing purpose calls
function test() {

    ++testCounter;
    console.log("-> testCounter = " + testCounter);

    var searchQuery = "look";

    if (testCounter == 1) {

        getCompatMode();
        wrappingSentencesWithinPTags();

        if (FolioPageFragment.getDirection() == "HORIZONTAL")
            initHorizontalDirection();

        highlightSearchResult(searchQuery, 1);

    } else if (testCounter == 2) {

        makeSearchResultsInvisible();

    } else if (testCounter == 3) {

        highlightSearchResult(searchQuery, 2);

    } else if (testCounter == 4) {

    }
}

function highlightSearchResult(searchQuery, occurrenceInChapter) {

    if (searchQuery == lastSearchQuery) {
        makeSearchResultsInvisible();
    } else {
        resetSearchResults();
        searchResults = applySearchResultClass(searchQuery);
        console.debug("-> Search Query Found = " + searchResults.length);
    }

    applySearchResultVisibleClass(occurrenceInChapter);
    FolioPageFragment.loadingFinished()
}

function applySearchResultClass(searchQuery) {

    var searchQueryRegExp = new RegExp(escapeRegExp(searchQuery), "i");

    var searchResults = [];
    var searchChildNodesArray = [];
    var elementArray = [];
    var textNodeArray = [];

    var bodyElement = document.getElementsByTagName('body')[0];
    var elementsInBody = bodyElement.getElementsByTagName('*');

    for (var i = 0 ; i < elementsInBody.length ; i++) {

        var childNodes = elementsInBody[i].childNodes;

        for (var j = 0; j < childNodes.length; j++) {

            if (childNodes[j].nodeType == Node.TEXT_NODE &&
                childNodes[j].nodeValue.trim().length) {
                //console.log("-> " + childNodes[j].nodeValue);

                if (childNodes[j].nodeValue.match(searchQueryRegExp)) {
                    //console.log("-> Found -> " + childNodes[j].nodeValue);

                    searchChildNodesArray.push(
                        getSearchChildNodes(childNodes[j].nodeValue, searchQuery));

                    elementArray.push(elementsInBody[i]);
                    textNodeArray.push(childNodes[j]);
                }
            }
        }
    }

    for (var i = 0 ; i < searchChildNodesArray.length ; i++) {

        var searchChildNodes = searchChildNodesArray[i];

        for (var j = 0 ; j < searchChildNodes.length ; j++) {

            if (searchChildNodes[j].className == "search-result")
                searchResults.push(searchChildNodes[j]);
            elementArray[i].insertBefore(searchChildNodes[j], textNodeArray[i]);
        }

        elementArray[i].removeChild(textNodeArray[i]);
    }

    lastSearchQuery = searchQuery;
    return searchResults;
}

function getSearchChildNodes(text, searchQuery) {

    var arrayIndex = [];
    var matchIndexStart = -1;
    var textChunk = "";
    var searchChildNodes = [];

    for (var i = 0, j = 0 ; i < text.length ; i++) {

        textChunk += text[i];

        if (text[i].match(new RegExp(escapeRegExp(searchQuery[j]), "i"))) {

            if (matchIndexStart == -1)
                matchIndexStart = i;

            if (searchQuery.length == j + 1) {

                var textNode = document.createTextNode(
                    textChunk.substring(0, textChunk.length - searchQuery.length));

                var searchNode = document.createElement("span");
                searchNode.className = "search-result";
                var queryTextNode = document.createTextNode(
                    text.substring(matchIndexStart, matchIndexStart + searchQuery.length));
                searchNode.appendChild(queryTextNode);

                searchChildNodes.push(textNode);
                searchChildNodes.push(searchNode);

                arrayIndex.push(matchIndexStart);
                matchIndexStart = -1;
                j = 0;
                textChunk = "";

            } else {
                j++;
            }

        } else {
            matchIndexStart = -1;
            j = 0;
        }
    }

    if (textChunk !== "") {
        var textNode = document.createTextNode(textChunk);
        searchChildNodes.push(textNode);
    }

    return searchChildNodes;
}

function makeSearchResultsVisible() {

    for (var i = 0 ; i < searchResults.length ; i++) {
        searchResults[i].className = "search-result-visible";
    }
    searchResultsInvisible = false;
}

function makeSearchResultsInvisible() {

    if (searchResultsInvisible)
        return;
    for (var i = 0 ; i < searchResults.length ; i++) {
        if (searchResults[i].className == "search-result-visible")
            searchResults[i].className = "search-result-invisible";
    }
    searchResultsInvisible = true;
}

function applySearchResultVisibleClass(occurrenceInChapter) {

    var searchResult = searchResults[occurrenceInChapter - 1];
    if (searchResult === undefined)
        return;
    searchResult.className = "search-result-visible";
    searchResultsInvisible = false;

    scrollToElement(searchResult);
}

function resetSearchResults() {

    for (var i = 0 ; i < searchResults.length ; i++) {
        searchResults[i].outerHTML = searchResults[i].innerHTML;
    }

    searchResults = [];
    lastSearchQuery = null;
    searchResultsInvisible = true;
}

function escapeRegExp(str) {
    return str.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1");
}

function scrollToLast() {
    var direction = FolioPageFragment.getDirection();
    var scrollingElement = bodyOrHtml();

    if (direction == "VERTICAL") {
        scrollingElement.scrollTop =
        scrollingElement.scrollHeight - document.documentElement.clientHeight;

    } else if (direction == "HORIZONTAL") {
        scrollingElement.scrollLeft =
        scrollingElement.scrollWidth - document.documentElement.clientWidth;
        WebViewPager.setPageToLast();
    }

    FolioPageFragment.loadingFinished()
}

function scrollToFirst() {

    var direction = FolioPageFragment.getDirection();
    var scrollingElement = bodyOrHtml();

    if (direction == "VERTICAL") {
        scrollingElement.scrollTop = 0;

    } else if (direction == "HORIZONTAL") {
        scrollingElement.scrollLeft = 0;
        WebViewPager.setPageToFirst();
    }

    FolioPageFragment.loadingFinished()
}


function scrollByPercentVertical(percent){
    if(percent < 0){
       return
    }
    var scrollingElement = bodyOrHtml()
    var scrollingElement = bodyOrHtml()
    scrollingElement.scrollTop = scrollingElement.scrollHeight * percent
}

function scrollByPercentHorizontal(percent){
    if(percent < 0){
       return
    }
    var scrollingElement = bodyOrHtml()
    var scrollingElement = bodyOrHtml()
    var clientWidth = document.documentElement.clientWidth;
    var dx = Math.round(scrollingElement.scrollWidth * percent);
    var offset = dx - (dx%clientWidth);
    scrollingElement.scrollLeft = offset;
    WebViewPager.setPageToPosition(Math.floor(dx/clientWidth));
}

function getCompatMode() {
    FolioWebView.setCompatMode(document.compatMode);
}

var scrollWidth;
var horizontalInterval;
var horizontalIntervalPeriod = 1000;
var horizontalIntervalCounter = 0;
var horizontalIntervalLimit = 3000;

function horizontalRecheck() {

    horizontalIntervalCounter += horizontalIntervalPeriod;

    if (window.scrollWidth != document.documentElement.scrollWidth) {
        postInitHorizontalDirection();
    }

    if (horizontalIntervalCounter >= horizontalIntervalLimit)
        clearInterval(horizontalInterval);
}

function initHorizontalDirection(scrollPercent) {

    var htmlElement = document.getElementsByTagName('html')[0];
    var bodyElement = document.getElementsByTagName('body')[0];
    var bodyStyle = bodyElement.currentStyle || window.getComputedStyle(bodyElement);
    var paddingRight = parseInt(bodyStyle.paddingRight, 10)
    console.log(paddingRight)
    var clientWidth = document.documentElement.clientWidth;
    var scrollWidth = document.documentElement.scrollWidth ;
    console.log(scrollWidth)
    if (scrollWidth > clientWidth
        && scrollWidth > document.documentElement.offsetWidth) {
        scrollWidth += paddingRight;
    }
    htmlElement.style.width = scrollWidth + 'px';
    var ratio = window.devicePixelRatio
    FolioPageFragment.loadingFinished(scrollWidth*ratio, clientWidth*ratio)


}



function preInitHorizontalDirection() {

    //console.log(window);
    //console.log("-> " + document.getElementsByTagName('title')[0].innerText);
    var htmlElement = document.getElementsByTagName('html')[0];
    var bodyElement = document.getElementsByTagName('body')[0];

    htmlElement.style.width = null;
    bodyElement.style.width = null;
    htmlElement.style.height = null;
    bodyElement.style.height = null;

    var bodyStyle = bodyElement.currentStyle || window.getComputedStyle(bodyElement);
    var paddingTop = parseInt(bodyStyle.paddingTop, 10);
    var paddingRight = parseInt(bodyStyle.paddingRight, 10);
    var paddingBottom = parseInt(bodyStyle.paddingBottom, 10);
    var paddingLeft = parseInt(bodyStyle.paddingLeft, 10);

    var pageWidth = document.documentElement.clientWidth - (paddingLeft + paddingRight);
    var pageHeight = document.documentElement.clientHeight - (paddingTop + paddingBottom);


    bodyElement.style.webkitColumnGap = (paddingLeft + paddingRight) + 'px';
    bodyElement.style.webkitColumnWidth = pageWidth + 'px';

    htmlElement.style.height = (pageHeight + (paddingTop + paddingBottom)) + 'px';
    bodyElement.style.height = pageHeight + 'px';
}

function postInitHorizontalDirection() {

    var htmlElement = document.getElementsByTagName('html')[0];
    var bodyElement = document.getElementsByTagName('body')[0];
    var bodyStyle = bodyElement.currentStyle || window.getComputedStyle(bodyElement);
    var paddingTop = parseInt(bodyStyle.paddingTop, 10);
    var paddingRight = parseInt(bodyStyle.paddingRight, 10);
    var paddingBottom = parseInt(bodyStyle.paddingBottom, 10);
    var paddingLeft = parseInt(bodyStyle.paddingLeft, 10);
    var clientWidth = document.documentElement.clientWidth;

    var scrollWidth = document.documentElement.scrollWidth;
    if (scrollWidth > clientWidth
        && scrollWidth > document.documentElement.offsetWidth) {
        scrollWidth += paddingRight;
    }
    var newBodyWidth = scrollWidth - (paddingLeft + paddingRight);
    window.scrollWidth = scrollWidth;

    htmlElement.style.width = scrollWidth + 'px';
    bodyElement.style.width = newBodyWidth + 'px';

    var pageCount = Math.round(scrollWidth / clientWidth);
    var pageCountFloat = scrollWidth / clientWidth;
    if(clientWidth === 0) {
      pageCount = 1
      pageCountFloat = 1
    }

    if (pageCount != pageCountFloat) {
        console.warn("-> pageCount = " + pageCount + ", pageCountFloat = " + pageCountFloat
            + ", Something wrong in pageCount calculation");
    }

    //console.log("-> scrollWidth = " + scrollWidth);
    //console.log("-> newBodyWidth = " + newBodyWidth);
    //console.log("-> pageCount = " + pageCount);

    FolioPageFragment.setHorizontalPageCount(pageCount);
}

/**
 Audio Mark ID - marks an element with an ID with the given class and scrolls to it
 */
function audioMarkID(className, id) {
    if (audioMarkClass)
        removeAllClasses(audioMarkClass);

    audioMarkClass = className
    var el = document.getElementById(id);

    scrollToElement(el);
    el.classList.add(className)
}

function setMediaOverlayStyle(style){
    document.documentElement.classList.remove("mediaOverlayStyle0", "mediaOverlayStyle1", "mediaOverlayStyle2")
    document.documentElement.classList.add(style)
}

function setMediaOverlayStyleColors(color, colorHighlight) {
    var stylesheet = document.styleSheets[document.styleSheets.length-1];
//    stylesheet.insertRule(".mediaOverlayStyle0 span.epub-media-overlay-playing { background: "+colorHighlight+" !important }")
//    stylesheet.insertRule(".mediaOverlayStyle1 span.epub-media-overlay-playing { border-color: "+color+" !important }")
//    stylesheet.insertRule(".mediaOverlayStyle2 span.epub-media-overlay-playing { color: "+color+" !important }")
}

var currentIndex = -1;


function findSentenceWithIDInView(els) {
    // @NOTE: is `span` too limiting?
    for(indx in els) {
        var element = els[indx];

        // Horizontal scroll
        if (document.body.scrollTop == 0) {
            var elLeft = document.body.clientWidth * Math.floor(element.offsetTop / window.innerHeight);
            // document.body.scrollLeft = elLeft;

            if (elLeft == document.body.scrollLeft) {
                currentIndex = indx;
                return element;
            }

        // Vertical
        } else if(element.offsetTop > document.body.scrollTop) {
            currentIndex = indx;
            return element;
        }
    }

    return null
}

function findNextSentenceInArray(els) {
    if(currentIndex >= 0) {
        currentIndex ++;
        return els[currentIndex];
    }

    return null
}

function resetCurrentSentenceIndex() {
    currentIndex = -1;
}

function rewindCurrentIndex() {
    currentIndex = currentIndex-1;
}

function getSentenceWithIndex(className) {
    var sentence;
    var sel = getSelection();
    var node = null;
    var elements = document.querySelectorAll("span.sentence");

    // Check for a selected text, if found start reading from it
    if (sel.toString() != "") {
        console.log(sel.anchorNode.parentNode);
        node = sel.anchorNode.parentNode;

        if (node.className == "sentence") {
            sentence = node

            for(var i = 0, len = elements.length; i < len; i++) {
                if (elements[i] === sentence) {
                    currentIndex = i;
                    break;
                }
            }
        } else {
            sentence = findSentenceWithIDInView(elements);
        }
    } else if (currentIndex < 0) {
        sentence = findSentenceWithIDInView(elements);
    } else {
        sentence = findNextSentenceInArray(elements);
    }

    var text = sentence.innerText || sentence.textContent;

    scrollToElement(sentence);

    if (audioMarkClass){
        removeAllClasses(audioMarkClass);
    }

    audioMarkClass = className;
    sentence.classList.add(className)
    return text;
}

function wrappingSentencesWithinPTags(){
    currentIndex = -1;
    "use strict";

    var rxOpen = new RegExp("<[^\\/].+?>"),
    rxClose = new RegExp("<\\/.+?>"),
    rxSupStart = new RegExp("^<sup\\b[^>]*>"),
    rxSupEnd = new RegExp("<\/sup>"),
    sentenceEnd = [],
    rxIndex;

    sentenceEnd.push(new RegExp("[^\\d][\\.!\\?]+"));
    sentenceEnd.push(new RegExp("(?=([^\\\"]*\\\"[^\\\"]*\\\")*[^\\\"]*?$)"));
    sentenceEnd.push(new RegExp("(?![^\\(]*?\\))"));
    sentenceEnd.push(new RegExp("(?![^\\[]*?\\])"));
    sentenceEnd.push(new RegExp("(?![^\\{]*?\\})"));
    sentenceEnd.push(new RegExp("(?![^\\|]*?\\|)"));
    sentenceEnd.push(new RegExp("(?![^\\\\]*?\\\\)"));
    //sentenceEnd.push(new RegExp("(?![^\\/.]*\\/)")); // all could be a problem, but this one is problematic

    rxIndex = new RegExp(sentenceEnd.reduce(function (previousValue, currentValue) {
                                            return previousValue + currentValue.source;
                                            }, ""));

    function indexSentenceEnd(html) {
        var index = html.search(rxIndex);

        if (index !== -1) {
            index += html.match(rxIndex)[0].length - 1;
        }

        return index;
    }

    function pushSpan(array, className, string, classNameOpt) {
        if (!string.match('[a-zA-Z0-9]+')) {
            array.push(string);
        } else {
            array.push('<span class="' + className + '">' + string + '</span>');
        }
    }

    function addSupToPrevious(html, array) {
        var sup = html.search(rxSupStart),
        end = 0,
        last;

        if (sup !== -1) {
            end = html.search(rxSupEnd);
            if (end !== -1) {
                last = array.pop();
                end = end + 6;
                array.push(last.slice(0, -7) + html.slice(0, end) + last.slice(-7));
            }
        }

        return html.slice(end);
    }

    function paragraphIsSentence(html, array) {
        var index = indexSentenceEnd(html);

        if (index === -1 || index === html.length) {
            pushSpan(array, "sentence", html, "paragraphIsSentence");
            html = "";
        }

        return html;
    }

    function paragraphNoMarkup(html, array) {
        var open = html.search(rxOpen),
        index = 0;

        if (open === -1) {
            index = indexSentenceEnd(html);
            if (index === -1) {
                index = html.length;
            }

            pushSpan(array, "sentence", html.slice(0, index += 1), "paragraphNoMarkup");
        }

        return html.slice(index);
    }

    function sentenceUncontained(html, array) {
        var open = html.search(rxOpen),
        index = 0,
        close;

        if (open !== -1) {
            index = indexSentenceEnd(html);
            if (index === -1) {
                index = html.length;
            }

            close = html.search(rxClose);
            if (index < open || index > close) {
                pushSpan(array, "sentence", html.slice(0, index += 1), "sentenceUncontained");
            } else {
                index = 0;
            }
        }

        return html.slice(index);
    }

    function sentenceContained(html, array) {
        var open = html.search(rxOpen),
        index = 0,
        close,
        count;

        if (open !== -1) {
            index = indexSentenceEnd(html);
            if (index === -1) {
                index = html.length;
            }

            close = html.search(rxClose);
            if (index > open && index < close) {
                count = html.match(rxClose)[0].length;
                pushSpan(array, "sentence", html.slice(0, close + count), "sentenceContained");
                index = close + count;
            } else {
                index = 0;
            }
        }

        return html.slice(index);
    }

    function anythingElse(html, array) {
        pushSpan(array, "sentence", html, "anythingElse");

        return "";
    }

    function guessSenetences() {
        var paragraphs = document.getElementsByTagName("p");

        Array.prototype.forEach.call(paragraphs, function (paragraph) {
            var html = paragraph.innerHTML,
                length = html.length,
                array = [],
                safety = 100;

            while (length && safety) {
                html = addSupToPrevious(html, array);
                if (html.length === length) {
                    if (html.length === length) {
                        html = paragraphIsSentence(html, array);
                        if (html.length === length) {
                            html = paragraphNoMarkup(html, array);
                            if (html.length === length) {
                                html = sentenceUncontained(html, array);
                                if (html.length === length) {
                                    html = sentenceContained(html, array);
                                    if (html.length === length) {
                                        html = anythingElse(html, array);
                                    }
                                }
                            }
                        }
                    }
                }

                length = html.length;
                safety -= 1;
            }

            try {
                paragraph.innerHTML = array.join("");
            } catch(err) {
                console.error(err);
                console.error("-> " + err.message);
            }
        });
    }

    guessSenetences();
}

function isElementVisible(element, isHorizontal) {

    var rect = element.getBoundingClientRect();

    if(isHorizontal)
        return rect.left > 0;
    else
        return rect.top > 0;
}

/**
Gets the first visible span from the displayed chapter and if it has id then usingId is true with
value as span id else usingId is false with value as span index. usingId and value is forwarded to
FolioPageFragment#storeFirstVisibleSpan(boolean, String) JavascriptInterface.

@param {boolean} isHorizontal - scrolling type of DirectionalViewpager#mDirection
*/
function getFirstVisibleSpan(isHorizontal) {

    var spanCollection = document.querySelectorAll("span.sentence");

    if (spanCollection.length == 0) {
        FolioPageFragment.storeFirstVisibleSpan(false, 0);
        return;
    }

    var spanIndex = 0;
    var spanElement;

    for (var i = 0 ; i < spanCollection.length ; i++) {
        if (isElementVisible(spanCollection[i], isHorizontal)) {
            spanIndex = i;
            spanElement = spanCollection[i];
            break;
        }
    }

    var usingId = spanElement.id ? true : false;
    var value = usingId ? spanElement.id : spanIndex;
    FolioPageFragment.storeFirstVisibleSpan(usingId, value);
}

/**
Scrolls the web page to particular span using id or index

@param {boolean} usingId - if span tag has id then true or else false
@param {number} value - if usingId true then span id else span index
*/
function scrollToSpan(usingId, value) {

    if (usingId) {
        var spanElement = document.getElementById(value);
        if (spanElement)
            scrollToElement(spanElement);
    } else {
        var spanCollection = document.querySelectorAll("span.sentence");;
        if (spanCollection.length == 0 || value < 0 || value >= spanCollection.length
            || value == null) {
            //LoadingView.hide();
            FolioPageFragment.loadingFinished()
            return;
        }
        scrollToElement(spanCollection[value]);
    }

    //LoadingView.hide();
    FolioPageFragment.loadingFinished()
}

// Class based onClick listener

function addClassBasedOnClickListener(schemeName, querySelector, attributeName, selectAll) {
    if (selectAll) {
        // Get all elements with the given query selector
        var elements = document.querySelectorAll(querySelector);
        for (elementIndex = 0; elementIndex < elements.length; elementIndex++) {
            var element = elements[elementIndex];
            addClassBasedOnClickListenerToElement(element, schemeName, attributeName);
        }
    } else {
        // Get the first element with the given query selector
        var element = document.querySelector(querySelector);
        addClassBasedOnClickListenerToElement(element, schemeName, attributeName);
    }
}

function addClassBasedOnClickListenerToElement(element, schemeName, attributeName) {
    // Get the content from the given attribute name
    var attributeContent = element.getAttribute(attributeName);
    // Add the on click logic
    element.setAttribute("onclick", "onClassBasedListenerClick(\"" + schemeName + "\", \"" + encodeURIComponent(attributeContent) + "\");");
}

var onClassBasedListenerClick = function(schemeName, attributeContent) {
    // Prevent the browser from performing the default on click behavior
    event.preventDefault();
    // Don't pass the click event to other elemtents
    event.stopPropagation();
    // Create parameters containing the click position inside the web view.
    var positionParameterString = "/clientX=" + event.clientX + "&clientY=" + event.clientY;
    // Set the custom link URL to the event
    window.location = schemeName + "://" + attributeContent + positionParameterString;
}

function getHighlightString(style) {
    var range = window.getSelection().getRangeAt(0);
    var selectionContents = range.extractContents();
    var elm = document.createElement("highlight");
    var id = guid();

    elm.appendChild(selectionContents);
    elm.setAttribute("id", id);
    elm.setAttribute("onclick","callHighlightURL(this);");
    elm.setAttribute("class", style);

    range.insertNode(elm);
    thisHighlight = elm;

    var params = [];
    params.push({id: id, rect: getRectForSelectedText(elm)});
    Highlight.getHighlightJson(JSON.stringify(params));
}

function goToHighlight(highlightId){
    var element = document.getElementById(highlightId.toString());
    if (element)
        scrollToElement(element);

    //LoadingView.hide();
    FolioPageFragment.loadingFinished()
}

function goToAnchor(anchorId) {
    var element = document.getElementById(anchorId);
    if (element)
        scrollToElement(element);

//    LoadingView.hide();
    FolioPageFragment.loadingFinished()
}

function array_diff(array1, array2){
    var difference = $.grep(array1, function(el) { return $.inArray(el,array2) < 0});
    return difference.concat($.grep(array2, function(el) { return $.inArray(el,array1) < 0}));;
}
